/*
 * PhysicsTest
 *
 * Written by Soothsayer in January 2010
 *
 * Website: http://www.cppanddelphi.atw.hu
 * Contact: cppanddelphi@gmail.com
 */
unit TPhysics;

interface

uses
  TVector, TMatrix;
type
  CNorm = array[0..3] of CVector;
  CPoly = array[0..4] of CVector;
  cpBB = record
    l: Single;
    b: Single;
    r: Single;
    t: Single;
  end;
type
//-------------------------------------------------------------------------------------------------------
// Class body
//-------------------------------------------------------------------------------------------------------
  CBody = class
  private
    Norm: CNorm;
    Size: CVector;
    Mass: Single;
    FI: Single;
    InvMass: Single;
    InvI: Single;
    BiasedVelocity: CVector;
    BiasedAngularVelocity: Single;
    Force: CVector;
    Torque: Single;
    BB: cpBB;
  public
    Poly: CPoly;
    Friction: SIngle;
    Position: CVector;
    Rotation: Single;
    Velocity: CVector;
    AngularVelocity: SIngle;

    constructor Create(const ASize: CVector; const AMass: Single); overload;
    procedure UpdateBody(var BB: CpBB);
  end;

type
  TContact = record
    Position: CVector;
    Normal: CVector;
    R1: CVector;
    R2: CVector;
    Separation: Single;
    Pn: Single;
    Pt: Single;
    Pnb: Single;
    MassNormal: Single;
    MassTangent: Single;
    Bias: Single;
  end;
var
  TempContacts: array of TContact;
type
  TContacts = array [0..10] of TContact;

type
//-------------------------------------------------------------------------------------------------------
// Class arbiter
//-------------------------------------------------------------------------------------------------------
  CArbiter = class
  private
    FBody1: CBody;
    FBody2: CBody;
    FFriction: Single;

    procedure PreStep(const InvDT: Single);
    procedure ApplyImpulse;
  public
    fPos: Integer;
    FNumContacts: Integer;
    FContacts: TContacts;

    constructor Create(const ABody1, ABody2: CBody; const AContacts: TContacts; const ANumContacts: Integer);
    destructor Destroy; override;
    procedure Update(const NewContacts: TContacts; const NewNumContacts: Integer);
    property  Body1: CBody read FBody1;
    property  Body2: CBody read FBody2;
  end;

  TListItem = record
    Body1: Byte;
    Body2: Byte;
  end;

var
  ArbiterCount: Integer;
  ArbiterList: array [0..16383] of TListItem;
  ArbiterBuckets: array [0..127, 0..127] of CArbiter;

type
//-------------------------------------------------------------------------------------------------------
// Class world
//-------------------------------------------------------------------------------------------------------
  CWorld = class
  private
    FGravity: CVector;
    FIterations: Integer;

    procedure BroadPhase;
  public
    FBodies: array of CBody;

    procedure Add(const Body: CBody); overload;
    procedure Clear;
    procedure Step(const DT: Single);
    constructor Create(const AGravity: CVector; const AIterations: Integer);
  end;

implementation

function min(a, b: Single): Single;
begin
  if a < b then Result := a
  else Result := b;
end;

function max(a, b: Single): Single;
begin
  if a > b then Result := a
  else Result := b;
end;

function Clamp(const A, Low, High: Single): Single;
begin
  Result := Max(Low,Min(A,High));
end;

function FindMSA(PolyI, PolyJ: CPoly; Norm: CNorm): TContact;
var
  c: TContact;
  i, j: Integer;
  dist, max, mindist: Single;
begin
  mindist := 1e10;
  for i := 0 to 3 do
  begin
    max := -1e10;
    for j := 0 to 3 do
    begin
      dist := dotV(SubV(PolyI[i], PolyJ[j]), Norm[i]);
      if (max < dist)
      then begin
        max := dist;
        c.Normal := Norm[i];
        c.Separation := dist;
      end;
    end;
    if mindist > max
    then begin
      mindist := max;
      Result := c;
    end;
  end;
end;

function PointInPoly(p: CVector; Poly: CPoly; Norm: CNorm): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to 3 do
    if dotV(SubV(Poly[i], p), norm[i]) < 0 then Exit;
  Result := True;
end;

procedure AddContact(c: TContact; V: CVector; Sign: Integer);
begin
  SetLength(TempContacts, Length(TempContacts) + 1);
  TempContacts[High(TempContacts)].Normal.X := c.Normal.X * Sign;
  TempContacts[High(TempContacts)].Normal.Y := c.Normal.Y * Sign;
  TempContacts[High(TempContacts)].Separation := c.Separation;
  TempContacts[High(TempContacts)].Position := V;
end;

procedure FindVerts(const BodyA, BodyB: CBody; c: TContact; Sign: Integer);
var
  i: Integer;
  V: CVector;
begin
  for i := 0 to 3 do
  begin
    V := BodyA.Poly[i];
    if PointInPoly(V, BodyB.Poly, BodyB.Norm) then AddContact(c, V, Sign);
    V := BodyB.Poly[i];
    if PointInPoly(V, BodyA.Poly, BodyA.Norm) then AddContact(c, V, Sign);
  end;
end;

function Collide(var Contacts: TContacts; const BodyA, BodyB: CBody): Integer;
var
  i: Integer;
  V: CVector;
  c0, c1: TContact;
begin
  Result := 0;
  SetLength(TempContacts, 0);

  c0 := FindMSA(BodyA.Poly, BodyB.Poly, BodyA.Norm);
  c1 := FindMSA(BodyB.Poly, BodyA.Poly, BodyB.Norm);
  if (c0.Separation < 0) and (c1.Separation < 0) then Exit;
  if c0.Separation < c1.Separation
  then
    FindVerts(BodyA, BodyB, c0, 1)
  else
    FindVerts(BodyA, BodyB, c1, -1);

  for i := 0 to High(TempContacts) do
  begin
    Contacts[Result] := TempContacts[i];
    Inc(Result);
  end;
end;




constructor CBody.Create(const ASize: CVector; const AMass: Single);
begin
  Friction := 0.2;
  Size := ASize;
  Mass := AMass;

  if (Mass < 3.4e+38 - 1e31) then begin
    InvMass := 1 / Mass;
    FI := Mass * ((Size.X * Size.X) + (Size.Y * Size.Y)) / 12;
    InvI := 1 / FI;
  end else begin
    InvMass := 0;
    FI := 3.4e+38;
    InvI := 0;
  end;
end;

procedure CBody.UpdateBody(var BB: CpBB);
var
  i: Integer;
  S, V: CVector;
  M: CMatrix;
begin
  S := MultiplyV(Size, 0.5);
  V := Position;
  M := CreateM(Rotation);
  Poly[0] := AddV(V, MultiplyM(M, CreateV( S.x, -S.y)));
  Poly[1] := AddV(V, MultiplyM(M, CreateV( S.x,  S.y)));
  Poly[2] := AddV(V, MultiplyM(M, CreateV(-S.x,  S.y)));
  Poly[3] := AddV(V, MultiplyM(M, CreateV(-S.x, -S.y)));
  Poly[4] := Poly[0];

  BB.r := Poly[0].x;
  BB.l := BB.r;
  BB.t := Poly[0].y;
  BB.b := BB.t;

  for i:= 0 to 3 do
  begin
    V := Poly[i];
    BB.l := min(BB.l, V.x);
    BB.r := max(BB.r, V.x);
    BB.b := min(BB.b, V.y);
    BB.t := max(BB.t, V.y);
  end;

  Norm[0] := CreateV(cos(Rotation), sin(Rotation));
  Norm[1] := CreateV(-Norm[0].Y, Norm[0].X);
  Norm[2] := NegativeV(Norm[0]);
  Norm[3] := NegativeV(Norm[1]);
end;




procedure CArbiter.PreStep(const InvDT: Single);
const
  BiasFactor = 0.2;
  AllowedPenetration = 0.1;
var
  RN1, RN2, kNormal, RT1, RT2, kTangent: Single;
  I: Integer;
  C: TContact;
  Tangent, P, V1, V2: CVector;
begin
  for I := 0 to FNumContacts - 1 do
  begin
    C := FContacts[I];

    C.R1 := SubV(C.Position, FBody1.Position);
    C.R2 := SubV(C.Position, FBody2.Position);

    kNormal := FBody1.InvMass + FBody2.InvMass;
    RN1 := Cross(C.R1, C.Normal);
    RN2 := Cross(C.R2, C.Normal);
    kNormal := kNormal + FBody1.InvI * RN1 * RN1 + FBody2.InvI * RN2 * RN2;
    C.MassNormal := 1 / kNormal;

    Tangent := Cross(C.Normal,1);
    RT1 := Cross(C.R1, Tangent);
    RT2 := Cross(C.R2, Tangent);
    kTangent := FBody1.InvMass + FBody2.InvMass;
    kTangent := kTangent + FBody1.InvI * RT1 * RT1 + FBody2.InvI * RT2 * RT2;
    C.MassTangent := 1 / kTangent;

    C.Bias := -BiasFactor * InvDT * Min(0, -C.Separation + AllowedPenetration);
    c.Pnb := 0;

    P := AddV(MultiplyV(C.Normal, C.Pn), MultiplyV(Tangent, C.Pt));
    FBody1.Velocity := SubV(FBody1.Velocity, MultiplyV(P, FBody1.InvMass));
    FBody1.AngularVelocity := FBody1.AngularVelocity - (FBody1.InvI * Cross(c.R1,P));
    FBody2.Velocity := AddV(FBody2.Velocity, MultiplyV(P, FBody2.InvMass));
    FBody2.AngularVelocity := FBody2.AngularVelocity + (FBody2.InvI * Cross(C.R2,P));

    FContacts[I] := C;
  end;
end;

procedure CArbiter.ApplyImpulse;
var
  I: Integer;
  C: TContact;
  j, t, v1, v2, vr, jb, vb1, vb2, n, r1, r2: CVector;
  jt, jtOld, jn, jnOld, vrt, vrn, vbn, jbn, jbnOld, jtMax: Single;
begin
  for I := 0 to FNumContacts - 1 do
  begin
    C := FContacts[I];
    n := C.Normal;
    r1 := C.R1;
    r2 := C.R2;

    vb1 := AddV(FBody1.BiasedVelocity, Cross(FBody1.BiasedAngularVelocity, r1));
    vb2 := AddV(FBody2.BiasedVelocity, Cross(FBody2.BiasedAngularVelocity, r2));
    vbn := dotV(subV(vb2, vb1), n);

    jbn := (c.bias - vbn) * c.MassNormal;
    jbnOld := c.Pnb;
    c.Pnb := max(jbnOld + jbn, 0.0);
    jbn := c.Pnb - jbnOld;

    jb := multiplyV(n, jbn);
    FBody1.BiasedVelocity := SubV(FBody1.BiasedVelocity, MultiplyV(jb, FBody1.InvMass));
    FBody1.BiasedAngularVelocity := FBody1.BiasedAngularVelocity - (FBody1.InvI * Cross(r1, jb));
    FBody2.BiasedVelocity := AddV(FBody2.BiasedVelocity, MultiplyV(jb, FBody2.InvMass));
    FBody2.BiasedAngularVelocity := FBody2.BiasedAngularVelocity + (FBody2.InvI * Cross(r2, jb));

    v1 := AddV(FBody1.Velocity, Cross(FBody1.AngularVelocity, r1));
    v2 := AddV(FBody2.Velocity, Cross(FBody2.AngularVelocity, r2));
    vr := subV(v2, v1);
    vrn := dotV(vr, n);

    jn := -vrn * C.MassNormal;
    jnOld := c.Pn;
    c.Pn := max(jnOld + jn, 0.0);
    jn := c.Pn - jnOld;

    t := Cross(n, 1);
    vrt := DotV(vr, t);

    jtMax := FFriction * C.Pn;
    jt := -vrt * C.MassTangent;
    jtOld := c.Pt;
    c.Pt := clamp(jtOld + jt, -jtMax, jtMax);
    jt := C.Pt - jtOld;

    j := addv(multiplyV(n, jn), multiplyV(t, jt));
    FBody1.Velocity := SubV(FBody1.Velocity, MultiplyV(j, FBody1.InvMass));
    FBody1.AngularVelocity := FBody1.AngularVelocity - (FBody1.InvI * Cross(r1, j));
    FBody2.Velocity := AddV(FBody2.Velocity, MultiplyV(j, FBody2.InvMass));
    FBody2.AngularVelocity := FBody2.AngularVelocity + (FBody2.InvI * Cross(r2, j));

    FContacts[I] := C;
  end;
end;

constructor CArbiter.Create(const ABody1, ABody2: CBody; const AContacts: TContacts; const ANumContacts: Integer);
var
  I: Integer;
begin
  FBody1 := ABody1;
  FBody2 := ABody2;

  FNumContacts := ANumContacts;
  for I := 0 to FNumContacts - 1 do
    FContacts[I] := AContacts[I];
  FFriction := Sqrt(FBody1.Friction * FBody2.Friction);
end;

destructor CArbiter.Destroy;
begin
  FBody1 := nil;
  FBody2 := nil;
end;

procedure CArbiter.Update(const NewContacts: TContacts; const NewNumContacts: Integer);
var
  I: Integer;
  C: TContact;
begin
  for I := 0 to NewNumContacts - 1 do
  begin
    C := NewContacts[I];
    C.Pn := FContacts[I].Pn;
    C.Pt := FContacts[I].Pt;
    C.Pnb := FContacts[I].Pnb;
    FContacts[I] := C;
  end;
  FNumContacts := NewNumContacts;
end;




procedure CWorld.BroadPhase;
const
  EPS = 1;
var
  I, J, k, NumContacts: Integer;
  dist: Single;
  BI, BJ: CBody;
  Arb: CArbiter;
  Contacts: TContacts;
begin
  FillChar(Contacts,SizeOf(Contacts),0);
  for i := 0 to High(FBodies) do
    FBodies[i].UpdateBody(FBodies[i].BB);
  for i := 0 to High(FBodies) - 1 do
  begin
    BI := FBodies[i];
    for j := I + 1 to High(FBodies) do
    begin
      BJ := FBodies[J];
      if (BI.InvMass <> 0) or (BJ.InvMass <> 0)
      then begin
      if ((BI.BB.l > BJ.BB.r + EPS) or (BJ.BB.l > BI.BB.r + EPS)) and
         ((BI.BB.t > BJ.BB.b + EPS) or (BJ.BB.t > BI.BB.b + EPS)) and (ArbiterBuckets[i, j] = nil) then continue;

        NumContacts := Collide(Contacts,BI,BJ);

        if (NumContacts > 0)
        then begin
          Arb := ArbiterBuckets[i, j];
          if Assigned(Arb) then
            CArbiter(Arb).Update(Contacts,NumContacts)
          else begin
            ArbiterBuckets[i, j]:= CArbiter.Create(BI,BJ,Contacts,NumContacts);
            Inc(ArbiterCount);
            ArbiterBuckets[i, j].fPos := ArbiterCount;
            ArbiterList[ArbiterCount].Body1 := i;
            ArbiterList[ArbiterCount].Body2 := j;
          end;
        end
        else begin
          if Assigned(ArbiterBuckets[i, j])
          then begin
            ArbiterBuckets[ArbiterList[ArbiterCount].Body1, ArbiterList[ArbiterCount].Body2].fPos := ArbiterBuckets[i, j].fPos;
            ArbiterList[ArbiterBuckets[i, j].fPos] := ArbiterList[ArbiterCount];
            Dec(ArbiterCount);
            ArbiterBuckets[i, j] := nil;
          end;
        end;
      end;
    end;
  end;
end;

procedure CWorld.Add(const Body: CBody);
begin
  SetLength(FBodies, Length(FBodies) + 1);
  FBodies[High(FBodies)] := Body;
end;

procedure CWorld.Clear;
begin
  ArbiterCount := -1;
  SetLength(FBodies, 0);
end;

constructor CWorld.Create(const AGravity: CVector; const AIterations: Integer);
begin
  FGravity := AGravity;
  FIterations := AIterations;
  SetLength(FBodies, 0);
  ArbiterCount := -1;
end;

procedure CWorld.Step(const DT: Single);
var
  InvDT: Single;
  I, J, k: Integer;
  B: CBody;
  A: CArbiter;
begin
  if (DT > 0) then InvDT := 1 / DT else InvDT := 0;

  for I := 0 to High(FBodies) do
  begin
    B := FBodies[i];
    if (B.InvMass <> 0)
    then begin
      B.Velocity := AddV(B.Velocity, MultiplyV(AddV(FGravity, MultiplyV(B.Force, B.InvMass)), DT));
      B.AngularVelocity := B.AngularVelocity + (DT * B.InvI * B.Torque);
      B.BiasedVelocity := CreateV(0,0);
      B.BiasedAngularVelocity := 0;
    end;
  end;

  BroadPhase;

  for I := 0 to ArbiterCount do
  begin
    A := ArbiterBuckets[ArbiterList[i].Body1, ArbiterList[i].Body2];
    A.PreStep(InvDT);
  end;

  for i := 0 to FIterations - 1 do
  begin
    for j := 0 to ArbiterCount do
    begin
      A := ArbiterBuckets[ArbiterList[j].Body1, ArbiterList[j].Body2];
      A.ApplyImpulse;
    end;
  end;

  for I := 0 to High(FBodies) do
  begin
    B := FBodies[i];
    B.Position := AddV(B.Position, MultiplyV(AddV(B.Velocity, B.BiasedVelocity), DT));
    B.Rotation := B.Rotation + (DT * (B.AngularVelocity + B.BiasedAngularVelocity));
    B.Force := CreateV(0,0);
    B.Torque := 0;
  end;
end;

end.

